/*
 * Copyright (c) 2000-2002 Silicon Graphics, Inc and LANL  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it would be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * Contact information:  Silicon Graphics, Inc., 1600 Amphitheatre Pkwy, 
 * Mountain View, CA  94043, or:
 * 
 * http://www.sgi.com 
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/sysmacros.h>
#include <sys/times.h>
#include <sys/utsname.h>
#include "csaacct.h"

#include <grp.h>
#include <malloc.h>
#include <memory.h>
#include <pwd.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "acctdef.h"
#include "acctmsg.h"
#include "csa_conf.h"
#include "sbu.h"

#include "csabuild.h"
#include "extern_build.h"

/*
 *	nqs.c:
 *		Routines to preprocess the NQS accounting data.  One record
 *		is written for each segment of a request.  Requests are
 *		uniquely identified by their sequence number/machine ID
 *		pair.  
 *
 *
 *		The output data is ordered by NQ_INIT time.  If a segment does
 *		not have a NQ_INIT record, NQ_RECV time is used instead.
 *
 *		NQS writes 7 types of accounting records.  For a given request,
 *		the order in which the records appear in the accounting file is:
 *
 *  NQ_RECV --> NQ_SENT --> NQ_INIT --> NQ_SPOOL --> NQ_TERM  -->  NQ_DISP
 *		               ^                       |
 *		               |                       |
 *		               -------------------------
 *
 *		The NQ_SENT records pertain to the pipeclient.  These records
 *		help tie the pipeclient process (which has uid = root) to the
 *		user's request.
 *			NQ_SENT_INIT --> NQ_SENT_TERM
 *
 *		The NQ_SPOOL records pertain to the netclient.  These records
 *		have a function similar to that of NQ_SENT records.  If the
 *		stdout and stderr destinations are different, there are 2
 *		pairs of NQ_SPOOL records.  If they are the same, there is
 *		only 1 pair.
 *			NQ_SPOOL_INIT --> NQ_SPOOL_TERM
 *
 *		The last type of record (NQ_INFO) is informational and contains
 *		no information about a user request.
 *
 *		A segment begins with either an NQ_RECV or NQ_INIT record and
 *		is terminated by an NQ_TERM record.  There are separate segments
 *		for the pipeclient (NQ_SENT) and netclient (NQ_SPOOL) processes.
 *
 *		A request begins with an NQ_RECV record and terminates with 
 *		an NQ_DISP record.
 */

#define NQS_NO_LTYPE		-1

/*
 *	Structures defining the linked list.
 *	Information is kept by request sequence number for
 *	all the RECV/INIT to TERM periods.  There is a separate
 *	nqsbs node for each such period.
 */
/*
 *	Segments merely point to a request's nqsbs list.
 */
struct	nqsseg {
	int		upind;		/* uptime index for this record */
	int		findex;		/* pacct file index for this rec */
	int64_t		pacct_offset;	/* pacct offset for this record */
	struct	nqsbs	*riptr;		/* nqsbs(c) nodes for this seqno */
	struct	nqsseg	*nptr;		/* next node for this seqno */
};

/*
 *	Structures containing running info about each nqs request.
 *	There is 1 structure per request sequence number.
 */
struct	nqsreq {
	int64_t		seqno;		/* request sequence number */
	int64_t		mid;		/* machine id number */
	time_t		stime;		/* RECV or INIT time stamp */
	time_t		sqtime;		/* start queue wait time */
	short		ltype;		/* previous rec type */
	short		lsubtype;	/* previous rec subtype */
	struct	nqsseg	*sfirst;	/* 1st nqs segment */
	struct	nqsseg	*slast;		/* last nqs segment */
	struct	nqsreq 	*bptr;		/* previous nqs request */
	struct	nqsseg	*ppfirst;	/* 1st routing/spooling segment */
	struct	nqsseg	*pplast;	/* last routing/spooling segment */
	struct	nqsreq	*nptr;		/* next nqs request */
};

/*
 *	Pointers to the request list
 */
struct req_head {
	struct	nqsreq	*first;		/* 1st node in the list */
	struct	nqsreq	*last;		/* last node in the list */
};

/*
 *	Global variables
 */
int	nout_rec = 0;			/* # output recs (riptr nodes) */
int	cur_findex = 0;
struct	req_head	*rhead;		/* pointer to the request list */
struct	nqs_track *tr_nqs;		/* global nqs request tracking struct */

char *Type_statname[] = {	/* names for NQS_type */
	"Unknown",	/* NQS_NONE = -2 --> 0 */
	"Info",		/* NQ_INFO = 1 */
	"Recv",		/* NQ_RECV = 2 */
	"Init",		/* NQ_INIT = 3 */
	"Term",		/* NQ_TERM = 4 */
	"Disp",		/* NQ_DISP = 5 */
	"Sent",		/* NQ_SENT = 6 */
	"Spool",	/* NQ_SPOOL = 7 */
	"Con",		/* NQ_CON = 8 */
};

char *Disp_statname[] = {	/* names for NQ_DISP */
	"Unknown",	/* NQS_NO_xxxx = -2 --> 0 */
	"Norm",		/* NQ_DISP_NORM = 1 */
	"Boot",		/* NQ_DISP_BOOT = 2 */
};

char *Info_statname[] = {	/* names for NQ_INFO */
	"Unknown",	/* NQ_NO_xxxx = -2 --> 0 */
	"AcctOn",	/* NQ_INFO_ACCTON = 1 */
	"AcctOff",	/* NQ_INFO_ACCTOFF = 2 */
};

char *Init_statname[] = {	/* names for NQ_INIT */
	"Unknown",	/* NQS_NO_INIT = -1 --> 0 */
	"Start",	/* NQ_INIT_START = 1 */
	"Restart",	/* NQ_INIT_RESTART = 2 */
	"Rerun",	/* NQ_INIT_RERUN = 3 */
	"I_Sent",	/* NQ_SENT_INIT = 4 */
	"I_Spool"	/* NQ_SPOOL_INIT = 5 */
};

char *Recv_statname[] = {	/* names for NQ_RECV */
	"Unknown",	/* NQS_NO_xxxx = -2 --> 0 */
	"New",		/* NQ_RECV_NEW = 1 */
	"Local",	/* NQ_RECV_LOCAL = 2 */
	"Remote",	/* NQ_RECV_REMOTE = 3 */
};

char *Term_statname[] = {	/* names for NQ_TERM */
	"Unknown",	/* NQS_NO_TERM = -2 --> 0 */
	"Exit",		/* NQ_TERM_EXIT = 1 */
	"Requeue",	/* NQ_TERM_REQUEUE = 2 */
	"Preempt",	/* NQ_TERM_PREEMPT = 3 */
	"Hold",		/* NQ_TERM_HOLD = 4 */
	"Oprerun",	/* NQ_TERM_OPRERUN = 5 */
	"T_Sent",	/* NQ_SENT_TERM = 6 */
	"T_Spool",	/* NQ_SPOOL_TERM = 7 */
	"Rerun"		/* NQ_TERM_RERUN = 8 */
};

/*
 *	Routine prototypes.
 */
static	void	dump_nqs_tbl();
static	struct	nqsreq	*find_request(int64_t, int64_t);
static	void	init_nqsbs(struct nqsbs *, struct nqsbs *);
static  int	jidcmp(const void *, const void *);
static	void	make_request(int);
static	struct	nqsseg	*make_segment(struct nqsreq *, int);
static	void	nqs_con(struct nqsbs *);
static	void	nqs_disp(struct nqsbs *);
static	void	nqs_init(struct nqsbs *);
static	void	nqs_recv(struct nqsbs *);
static	void	nqs_sent(struct nqsbs *);
static	void	nqs_spool(struct nqsbs *);
static	void	nqs_term(struct nqsbs *);
static	void	prc_nqs(struct nqsbs *);


/*
 *	process_nqs() - process NQS daemon record.
 */
int
process_nqs(int fd, int findex)
{
	static	int	tso_flag;
	static	time_t	lasttime = 0;
	time_t	btime;
	uint64_t jobid;

/*
 *	Check the record time information against the uptime info.
 */
	btime = acctent.nbs->time;
	if (lasttime) {
		if (lasttime > btime) {
			/*
			 *	If time stamps are out of order, keep on going.
			 *	Assume that the records belong in the current
			 *	or later up periods.
			 */
			if (!tso_flag) {
				tso_flag++;
				acct_err(ACCT_WARN,
				       _("The %s records were out of time order in file '%s%d'\n   near offset %lld due to a date change or an invalid file."),
					"nqs",
					pacct, findex, Pacct_offset);
			}
		}

	} else {
		if (btime < uptime[0].up_start) {
		    if (btime < EARLY_FUDGE) {
			char btime_str[26];
			time_t	time = EARLY_FUDGE;

			strcpy(btime_str, ctime(&btime));
			Nerror("REC_NQS: EARLY_FUDGE error (%d - %24.24s) ",
				btime, btime_str);
			Nerror("< (%d - %24.24s)\n", time, ctime(&time));
			acct_err(ACCT_WARN,
			    _("An earlier than expected time stamp was discovered in the file '%s%d' near offset (%lld).  Examine the data set."),
			      pacct, findex, Pacct_offset);

			if (i_opt) {
				acct_err(ACCT_INFO, 
					_("The record was ignored."));
				if (db_flag > 2) {
					Dump_rec(acctent.nbs,
						sizeof(struct nqsbs));
				}
				return(RC_BAD);
			}
		    }
		    if (upind != 0) {
				if (db_flag > 2) {
					Dump_rec(acctent.nbs,
						sizeof(struct nqsbs));
				}
				acct_err(ACCT_ABORT,
				       _("An earlier than expected time stamp was discovered in the file '%s%d' near offset (%lld).  Examine the data set."),
					pacct, findex, Pacct_offset);

		    } else {
				Shift_uptime(REC_NQS, pacct, findex);
		    }
		}
	}

	lasttime = btime;
	while(uptime[upind].up_stop < btime) {
		if (db_flag > 3) {
			Ndebug("process_nqs(4): Past stop time(%d) going "
				"to next NQS uptime(%d) - %s%d.\n",
				uptime[upind].up_stop, upind+1, pacct, findex);
		}
		upind++;
		if (upind > upIndex) {
			break;
		}
	}
	cur_findex = findex;

/*
 *	Process the NQS records - consolidate all associated data. Since
 *	there aren't valid jids in some of the NQS daemon records,
 *	don't attach the consolidated record to a segment now.  This
 *	will be done later by a call to add_segs_nqs() after all of the 
 *	pacct records have been processed.  The NQS consolidated records
 *	at that time have accurate information.
 */
	prc_nqs(acctent.nbs);

	return(RC_GOOD);
}


/*
 *	print_nqs_rec - print raw nqs requests.
 */
static	void
print_nqs_rec(struct nqsbs *nbuf)
{
	time_t	stime;

	Ndebug("   NQS_REC: type(%d), uid(%d-%s), jid(0x%llx).\n",
		nbuf->type, nbuf->uid, uid_to_name(nbuf->uid),
		nbuf->jid);

	Ndebug("   NQS_REC: ash(0x%llx), prid(%lld-%s).\n",
		nbuf->ash, nbuf->prid, prid_to_name(nbuf->prid));

	stime = nbuf->time;
	Ndebug("   NQS_REC: seqno(%lld), mid(%lld), time: %s",
		nbuf->seqno, nbuf->mid, ctime(&stime));

	switch(nbuf->type) {

	case NQ_INFO:
	    {
		int	subtype = nbuf->subtype < 0 ? 0 : nbuf->subtype;

		Ndebug("   NQ_INFO: subtype(%d - %s).\n", subtype,
			Info_statname[subtype]);
	    }
	    break;

	case NQ_SENT:
	    {
		int	subtype = nbuf->subtype < 0 ? 0 : nbuf->subtype;

		if (subtype == NQ_SENT_INIT) {
			Ndebug("   NQ_SENT: subtype(%d - %s).\n", subtype,
				Init_statname[subtype]);
		} else {
			Ndebug("   NQ_SENT: subtype(%d - %s).\n", subtype,
				Term_statname[subtype]);
		}
		Ndebug("   NQ_SENT: machname(%s), reqname(%s), quename(%s), "
			"qtype(%d)\n", nbuf->machname, nbuf->reqname,
			nbuf->quename, nbuf->qtype);
		Ndebug("   NQ_SENT: code(%lld), utime(%lld), stime(%lld)\n",
			nbuf->code, nbuf->utime, nbuf->stime);
	    }
	    break;

	case NQ_RECV:
	    {
		int	subtype = nbuf->subtype < 0 ? 0 : nbuf->subtype;

		Ndebug("   NQ_RECV: subtype(%d - %s).\n", subtype,
			Recv_statname[subtype]);
		Ndebug("   NQ_RECV: machname(%s), reqname(%s), quename(%s), "
			"qtype(%d)\n", nbuf->machname, nbuf->reqname,
			nbuf->quename, nbuf->qtype);
	    }
	    break;

	case NQ_INIT:
	    {
		int	subtype = nbuf->subtype < 0 ? 0 : nbuf->subtype;

		Ndebug("   NQ_INIT: subtype(%d - %s).\n", subtype,
			Init_statname[subtype]);
		Ndebug("   NQ_INIT: machname(%s), reqname(%s), quename(%s), "
			"qtype(%d)\n", nbuf->machname, nbuf->reqname,
			nbuf->quename, nbuf->qtype);
	    }
	    break;

	case NQ_TERM:
	    {
		int	subtype = nbuf->subtype < 0 ? 0 : nbuf->subtype;

		Ndebug("   NQ_TERM: subtype(%d - %s).\n", subtype,
			Term_statname[subtype]);
		Ndebug("   NQ_TERM: machname(%s), reqname(%s), quename(%s), "
			"qtype(%d)\n", nbuf->machname, nbuf->reqname,
			nbuf->quename, nbuf->qtype);
		Ndebug("   NQ_TERM: code(%lld), utime(%lld), stime(%lld)\n",
			nbuf->code, nbuf->utime, nbuf->stime);
	    }
	    break;

	case NQ_SPOOL:
	    {
		int	subtype = nbuf->subtype < 0 ? 0 : nbuf->subtype;

		if (subtype == NQ_SPOOL_INIT) {
			Ndebug("   NQ_SPOOL: subtype(%d - %s).\n", subtype,
				Init_statname[subtype]);
		} else {
			Ndebug("   NQ_SPOOL: subtype(%d - %s).\n", subtype,
				Term_statname[subtype]);
		}
		Ndebug("   NQ_SPOOL: machname(%s), reqname(%s), quename(%s), "
			"qtype(%d)\n", nbuf->machname, nbuf->reqname,
			nbuf->quename, nbuf->qtype);
	    }
	    break;

	case NQ_DISP:
	    {
		int	subtype = nbuf->subtype < 0 ? 0 : nbuf->subtype;

		Ndebug("   NQ_DISP: subtype(%d - %s).\n", subtype,
			Disp_statname[subtype]);
		Ndebug("   NQ_DISP: machname(%s), reqname(%s), quename(%s), "
			"qtype(%d)\n", nbuf->machname, nbuf->reqname,
			nbuf->quename, nbuf->qtype);
	    }
	    break;

	case NQ_CON:
	    {
		int	subtype;

		Ndebug("   NQ_CON: machname(%s), reqname(%s), quename(%s), "
			"qtype(%d)\n", nbuf->machname, nbuf->reqname,
			nbuf->quename, nbuf->qtype);
		Ndebug("   NQ_CON: code(%lld), utime(%lld), stime(%lld)\n",
			nbuf->code, nbuf->utime, nbuf->stime);
		subtype = nbuf->subtype < 0 ? 0 : nbuf->subtype;
		Ndebug("   NQ_CON: init_type(%d - %s).\n", nbuf->subtype,
			Init_statname[subtype]);
		subtype = nbuf->term_subtype < 0 ? 0 : nbuf->term_subtype;
		Ndebug("   NQ_CON: term_type(%d - %s).\n", nbuf->term_subtype,
			Term_statname[subtype]);
		subtype = nbuf->disp_subtype < 0 ? 0 : nbuf->disp_subtype;
		Ndebug("   NQ_CON: disp_type(%d - %s).\n", nbuf->disp_subtype,
			Disp_statname[subtype]);
		Ndebug("   NQ_CON: qwtime(%lld), tmtime(%d)\n",
			nbuf->qwtime, nbuf->tmtime);
	    }
	    break;

	default:
		Ndebug(" Unknown NQS request type(%d)\n", nbuf->type);

	}		/* end of switch(type) */

	return;
}


/*
 *	prc_nqs - process the raw nqs requests.
 */
void
prc_nqs(struct nqsbs *nbuf)
{
	static	int	nrec = 0;

	nrec++;
	if (db_flag > 8) {
		print_nqs_rec(nbuf);
	}

	switch(nbuf->type) {

	/*
	 *	Informational record 
	 */
	case NQ_INFO:
		if (db_flag > 8) {
			Ndebug("prc_nqs(9): INFO.\n");
		}
		break;		/* ignore this record?? */

	/*
	 *	A request has been routed (pipeclient).
	 *	If needed, add a new segment node to the end
	 *	of the linked list and then initialize the node.
	 */
	case NQ_SENT:
		nqs_sent(nbuf);
		break;

	/*
	 * 	A request has been received.
	 *	Add a request node to the end of the linked 
	 *	list and then initialize the node.
	 */
	case NQ_RECV:
		nqs_recv(nbuf);
		break;

	/*
	 *	A request has been initiated.
	 *	Add a new segment node to the end of this 
	 *	request's segment list.
	 */
	case NQ_INIT:
		nqs_init(nbuf);
		break;

	/*
	 *	A request has been terminated.
	 *	Add the termination code to the nqsbs node. 
	 *	Also, restart the start queue wait time.
	 */
	case NQ_TERM:
		nqs_term(nbuf);
		break;

	/*
	 *	A request's output has been spooled (netclient).
	 *	If necessary, add a request node to the end of
	 *	the linked list and then initialize the node.
	 */
	case NQ_SPOOL:
		nqs_spool(nbuf);
		break;

	/*
	 *	A request has been disposed - gone from the
	 *	system.	  Merely set the last record type.
	 */
	case NQ_DISP:
		nqs_disp(nbuf);
		break;

	/*
	 *	A recycled consolidated request has been found.
	 *	Save all data in consolidated structure.
	 */
	case NQ_CON:
		nqs_con(nbuf);
		break;

	default:
		acct_err(ACCT_CAUT,
		       _("An unknown record type (%4o) was found in the '%s' routine."),
			nbuf->type, "prc_nqs()");

	}		/* end of switch(type) */

	if (db_flag > 8) {
		dump_nqs_tbl();
	}

	return;
}


/*
 *	Find the request node that matches both the sequence number
 *	and machine ID by searching backwards through the request
 *	list.  
 *
 *	Return the pointer to the request node, if found.  Otherwise,
 *	return NULL.
 */
static	struct	nqsreq *
find_request(int64_t seqno, int64_t mid)
{
	struct	nqsreq	*req_ptr;

	req_ptr = rhead->last;
	while (req_ptr != NULL) {
		if ((req_ptr->seqno == seqno) && (req_ptr->mid == mid) ) {
			return(req_ptr);	/* return request node */

		} else {
			req_ptr = req_ptr->bptr;
		}
	}

	return(NULL);		/* request node not found */
}


/*
 *	Create the head pointer to the request list
 */
void
init_nqs()
{
	if ((rhead = (struct req_head *)malloc(sizeof(struct req_head)))
			== NULL) {
		acct_perr(ACCT_ABORT, errno,
			_("There was insufficient memory available when allocating '%s'."),
			"request head");
	}
	rhead->first = rhead->last = NULL;

	return;
}


/*
 *	Initialize a nqsbs(consolidated) node using information from the
 *	current input record if we're sure it's really there.
 */
static void
init_nqsbs(struct nqsbs *riptr, struct nqsbs *nbuf)
{
	riptr->start_time = nbuf->time;
	riptr->uid = nbuf->uid;
	if (db_flag > 7) {
		Ndebug("init_nqsbs(8): nbuf->uid(%d), riptr->uid(%d).\n",
			nbuf->uid, riptr->uid);
	}
	riptr->type	= NQ_CON;
	riptr->time	= nbuf->time;
	riptr->prid	= nbuf->prid;
	riptr->ash	= nbuf->ash;
	riptr->jid	= nbuf->jid;
	riptr->seqno	= nbuf->seqno;
	riptr->mid	= nbuf->mid;
	riptr->qtype	= nbuf->qtype;
	riptr->code	= 0;
	riptr->utime	= nbuf->utime;
	riptr->stime	= nbuf->stime;

	riptr->disp_subtype = NQS_NO_DISP;
	riptr->init_subtype = NQS_NO_INIT;
	riptr->term_subtype = NQS_NO_TERM;

	riptr->qwtime	= DEFAULT_QWTIME;
	riptr->tmtime	= 0;
	riptr->sbu_wt	= 1.0;

	strcpy(riptr->machname, nbuf->machname);
	strcpy(riptr->quename, nbuf->quename);
	strcpy(riptr->reqname, nbuf->reqname);

	return;
}

/*
 *	Add a request node (and its associated segment and nqsbs(c) nodes)
 *	to the end of the linked list
 */
static void
make_request(int typ)
{
	struct	nqsreq	*nw_req;

	/*
	 *	Allocate space for the new request node and zero it out
	 */
	if ((nw_req = (struct nqsreq *)malloc(sizeof(struct nqsreq)))
			== NULL) {
		acct_perr(ACCT_ABORT, errno,
			_("There was insufficient memory available when allocating '%s'."),
			"nqsreq node");
	}
	memset((char *)nw_req, '\0', sizeof(struct nqsreq));

	/*
	 *	Link the new node to the end of the list
	 */

	nw_req->nptr = NULL;
	if (rhead->last != NULL) {
		rhead->last->nptr = nw_req;
		nw_req->bptr = rhead->last;
		rhead->last = nw_req;

	} else {
		rhead->first = rhead->last = nw_req;
		nw_req->bptr = NULL;
	}

	/*
	 *	Now add the segment and nqsbs nodes
	 */
	nw_req->sfirst = nw_req->slast = NULL;
	nw_req->ppfirst = nw_req->pplast = NULL;
	make_segment(nw_req, typ);
	nw_req->ltype = NQS_NO_LTYPE;

	return;
}

/*
 *	Add a segment node and its associated reginfo node to
 *	the end of a request's segment list.  Return a pointer
 *	to the new segment node.
 */
static	struct	nqsseg *
make_segment(struct nqsreq *req_ptr, int typ)
{
	struct	nqsbs	*nw_nqs;
	struct	nqsseg	*nw_seg;

	/*
	 *	Allocate memory for the new nodes and initialize fields
	 */
	if ((nw_nqs = (struct nqsbs *)malloc(sizeof(struct nqsbs))) == NULL) {
		acct_perr(ACCT_ABORT, errno,
			_("There was insufficient memory available when allocating '%s'."),
			"nqsbs(c) node");
	}
	memset((char *)nw_nqs, '\0', sizeof(struct nqsbs));
	if (create_hdr1(&nw_nqs->hdr, ACCT_DAEMON_NQS) ) {
		acct_err(ACCT_WARN,
		       _("An error was returned from the call to the '%s' routine."),
			"create_hdr()");
	}
	nout_rec++;
	if (db_flag > 7) {
		Ndebug("make_segment(8): nout_rec(%d).\n", nout_rec);
	}
	if ((nw_seg = (struct nqsseg *)malloc(sizeof(struct nqsseg))) == NULL) {
		acct_perr(ACCT_ABORT, errno,
			_("There was insufficient memory available when allocating '%s'."),
			"segment node");
	}

	/*
	 *	Add the node to the linked list
	 */
	if (typ == NORM_REQ) {
		if (req_ptr->sfirst == NULL) {
			req_ptr->sfirst = req_ptr->slast = nw_seg;
		} else {
			req_ptr->slast->nptr = nw_seg;
			req_ptr->slast = nw_seg;
		}

	} else {	/* typ == AUX_REQ */
		if (req_ptr->ppfirst == NULL) {
			req_ptr->ppfirst = req_ptr->pplast = nw_seg;
		} else {
			req_ptr->pplast->nptr = nw_seg;
			req_ptr->pplast = nw_seg;
		}
	}

	/*
	 *	Initialize the new nodes with default values and
	 *	some info for the current location of this record in
	 *	the pacct file for possible error handling later
	 */
	nw_seg->upind = upind;
	nw_seg->findex = cur_findex;
	nw_seg->pacct_offset = Pacct_offset;
	nw_seg->nptr = NULL;
	nw_seg->riptr = nw_nqs;

	nw_nqs->code = NQS_NO_TERM;		/* code is on NQ_TERM rec */
	nw_nqs->term_subtype = NQS_NO_TERM;
	nw_nqs->qwtime = DEFAULT_QWTIME;
	nw_nqs->disp_subtype = NQS_NO_DISP;

	return(nw_seg);
}


/*
 *	Process NQ_CON records
 *	  If there isn't already a request node for this seqno.mid, make one.
 *
 */
void
nqs_con(struct nqsbs *nbuf)
{
	struct	nqsbs	*riptr = NULL;
	struct	nqsreq	*rqptr;

	if (db_flag > 8) {
		Ndebug("nqs_con(9): CON - seqno(%lld), mid(%lld).\n",
			nbuf->seqno, nbuf->mid);
	}

	if ((rqptr = find_request(nbuf->seqno, nbuf->mid)) == NULL) {
		/*
		 * This is a new request.  No other records for this
		 * request have been seen.
		 */
		if (db_flag > 2) {
			Ndebug("nqs_con(3): Only NQ_CON record "
				"found for seqno(%lld), subtype(%s).\n",
				nbuf->seqno, nbuf->subtype); 
		}
		make_request(NORM_REQ);
		rqptr = rhead->last;
		rqptr->seqno = nbuf->seqno;
		rqptr->mid = nbuf->mid;
		rqptr->stime = nbuf->time;
		riptr = rqptr->slast->riptr;
		init_nqsbs(riptr, nbuf);
	}

	return;
}


/*
 *	Process NQ_DISP records
 *	  Set the ltype and lsubtype in the request node and the dispose subtype
 *	  in the nqsbs node.  If there isn't already a request node for this
 *	  seqno.mid, make one.
 *
 *	  If this is a null (rejected) NQS job, the queue wait time is
 *	  the time between the NQ_RECV and NQ_DISP records.  Null jobs
 *	  have only NQ_RECV and NQ_DISP records.
 *
 *	  Note: NQ_DISP records do not contain valid jids.
 */
void
nqs_disp(struct nqsbs *nbuf)
{
	struct	nqsbs	*riptr = NULL;
	struct	nqsreq	*rqptr;
	struct	nqsseg	*sptr;

	if (db_flag > 8) {
		Ndebug("nqs_disp(9): DISP - seqno(%lld), mid(%lld).\n",
			nbuf->seqno, nbuf->mid);
	}

	if ((rqptr = find_request(nbuf->seqno, nbuf->mid)) == NULL) {
		/*
		 * This is a new request.  No other records for this
		 * request have been seen.
		 */
		if (db_flag > 2) {
			Ndebug("nqs_disp(3): Only NQ_DISP record "
				"found for seqno(%lld), subtype(%s).\n",
				nbuf->seqno, nbuf->subtype);
		}
		make_request(NORM_REQ);
		rqptr = rhead->last;
		rqptr->seqno = nbuf->seqno;
		rqptr->mid = nbuf->mid;
		rqptr->stime = nbuf->time;
		riptr = rqptr->slast->riptr;
		init_nqsbs(riptr, nbuf);
		riptr->jid = NQS_NO_JID;

	} else if (rqptr->slast == NULL) {
		/*
		 * This is a request that has only AUX_REQ records.
		 * Set the time stamp on the nqsbs(c) node to be that
		 * of the first AUX_REQ nqsbs(c) node.  This time will
		 * be closer to the time of the qsub.
		 */
		make_segment(rqptr, NORM_REQ);
		riptr = rqptr->slast->riptr;
		init_nqsbs(riptr, nbuf);
		riptr->jid = NQS_NO_JID;
		if (rqptr->ppfirst != NULL) {
			riptr->start_time = rqptr->ppfirst->riptr->start_time;
		}

	} else {
		/*
		 * This request has other NORM_REQ records.
		 */
		riptr = rqptr->slast->riptr;
	}

	/*
	 *	Figure out the queue wait time for null jobs.
	 */
	if (riptr->jid == NQS_NO_JID && rqptr->ltype == NQ_RECV) {
		riptr->qwtime = nbuf->time - rqptr->sqtime;
	}

	rqptr->ltype    = nbuf->type;
	rqptr->lsubtype = nbuf->subtype;

	sptr = rqptr->sfirst;
	while (sptr != NULL) {
		sptr->riptr->disp_subtype = nbuf->subtype;
		sptr = sptr->nptr;
	}

	return;
}


/*
 *	Process NQ_INIT records
 */
void
nqs_init(struct nqsbs *nbuf)
{
	struct nqsbs	*riptr;
	struct nqsreq	*rqptr;
	struct nqsseg	*sptr;

	if (db_flag > 8) {
		Ndebug("nqs_init(9): INIT - seqno(%lld), mid(%lld).\n",
			nbuf->seqno, nbuf->mid);
	}

	/*
	 *	Ideally, this seqno should already have a request node,
	 *	but if it doesn't, we'll make one
	 */
	if ((rqptr = find_request(nbuf->seqno, nbuf->mid)) == NULL) {
		if (db_flag > 2) {
			Ndebug("nqs_init(3): Making request node for "
				"seqno(%lld), type NQ_INIT, subtype(%s).\n",
				nbuf->seqno, nbuf->subtype);
		}
		make_request(NORM_REQ);
		rqptr = rhead->last;
		rqptr->seqno = nbuf->seqno;
		rqptr->mid = nbuf->mid;

	} else if (rqptr->slast == NULL) {
		make_segment(rqptr, NORM_REQ);
	}

	/*
	 *	If the previous record type for this seqno is not
	 *	NQ_RECV, we need to add a segment node to the end
	 *	of this seqno's list.
	 */
	if (rqptr->ltype != NQ_RECV && rqptr->ltype != NQS_NO_LTYPE) {
		sptr = make_segment(rqptr, NORM_REQ);
		riptr = sptr->riptr;
	} else {
		riptr = rqptr->slast->riptr;
	}

	/*
	 *	Initialize the nqsbs(c) node with values from the
	 *	input record
	 */
	init_nqsbs(riptr, nbuf);
	riptr->init_subtype = nbuf->subtype;
	if ((rqptr->ltype == NQ_RECV) || (rqptr->ltype == NQ_TERM) ) {
		riptr->qwtime = nbuf->time - rqptr->sqtime;
	}

	if (rqptr->ltype == NQ_RECV) {
		riptr->start_time = rqptr->stime;
	} else {
		riptr->start_time = nbuf->time;
	}

	rqptr->ltype    = nbuf->type;
	rqptr->lsubtype = nbuf->subtype;

	/*
	 *	Time stamp on output record is always the time stamp
	 *	of the NQ_INIT record, if it exists.
	 */
	rqptr->stime = riptr->start_time = nbuf->time;

	return;
}


/*
 *	Process NQ_RECV records
 *	  If a request node does not exist for the sequence number,
 *	  add one to the end of the request linked list.  Update
 *	  necessary fields.
 *
 *	  If there are multiple NQ_RECV records (request moves from pipe
 *	  to batch queue), the start queue wait time (sqtime) is set to the
 *	  time of the first record (usually pipe queue).
 *	  Also update the queue type and name so
 *	  we always have the most current type and name.
 *
 *	  The initial time stamp for this segment is set to the time of
 *	  the most recent NQ_RECV record.
 *
 *	  NOTE: NQ_RECV records do not contain valid jids.
 */
void
nqs_recv(struct nqsbs *nbuf)
{
	struct nqsbs	*riptr;
	struct nqsreq	*rqptr;

	if (db_flag > 8) {
		Ndebug("nqs_recv(9): RECV - seqno(%lld), mid(%lld).\n",
			nbuf->seqno, nbuf->mid);
	}

	if ((rqptr = find_request(nbuf->seqno, nbuf->mid)) == NULL) {
		make_request(NORM_REQ);
		rqptr = rhead->last;	
		rqptr->seqno = nbuf->seqno;
		rqptr->mid = nbuf->mid;
		rqptr->sqtime = nbuf->time;
		rqptr->stime = nbuf->time;
		riptr = rqptr->slast->riptr;
		init_nqsbs(riptr, nbuf);
		riptr->jid = NQS_NO_JID;
	} else if (rqptr->slast == NULL) {
		rqptr->sqtime = nbuf->time;	/* update */
		rqptr->stime = nbuf->time;	/* update */
		make_segment(rqptr, NORM_REQ);
		riptr = rqptr->slast->riptr;
		init_nqsbs(riptr, nbuf);
		riptr->jid = NQS_NO_JID;

	} else {
		riptr = rqptr->slast->riptr;

	}

	riptr->qtype = nbuf->qtype;
	strcpy(riptr->quename, nbuf->quename);

	rqptr->ltype    = nbuf->type;
	rqptr->lsubtype = nbuf->subtype;

	return;
}


/*
 *	Process NQ_SENT records
 *	  If a request node does not exist for the sequence number,
 *	  add one to the end of the request linked list.  The segment
 *	  node should be on the pp list not on the normal segment list.
 *
 *	  There should be a separate segment and nqsbs(c) node for each
 *	  NQ_SENT_INIT/NQ_SENT_TERM pair.
 *
 */
void
nqs_sent(struct nqsbs *nbuf)
{
	struct	nqsbs	*riptr;
	struct	nqsreq	*rqptr;

	if (db_flag > 8) {
		Ndebug("nqs_sent(9): SENT - seqno(%lld), mid(%lld).\n",
			nbuf->seqno, nbuf->mid);
	}

	if ((rqptr = find_request(nbuf->seqno, nbuf->mid)) == NULL) {
		/*
		 * This is a new request.  No other records for this
		 * request have been seen.
		 */
		if (db_flag >= 2) {
			Ndebug("nqs_sent(3): type NQ_SENT, subtype(%d).\n",
				nbuf->subtype);
		}
		make_request(AUX_REQ);
		rqptr = rhead->last;	
		rqptr->seqno = nbuf->seqno;
		rqptr->mid = nbuf->mid;
		riptr = rqptr->pplast->riptr;
		init_nqsbs(riptr, nbuf);

	} else if (rqptr->pplast == NULL) {
		/*
		 * Only NORM_REQ records have been seen for this request.
		 */
		if (db_flag > 2) {
			Ndebug("nqs_sent(3): pplast == NULL.\n");
		}
		make_segment(rqptr, AUX_REQ);
		riptr = rqptr->pplast->riptr;
		init_nqsbs(riptr, nbuf);
		/*
		 * preserve original qsub-time
		 */
		riptr->start_time = rqptr->stime;

	} else {
		/*
		 * This request has other AUX_REQ records.
		 */
		riptr = rqptr->pplast->riptr;
		if (db_flag > 2) {
			Ndebug("nqs_sent(3): examining segment node\n"
				"   init_subtype(%d), term_subtype(%d).\n",
				riptr->init_subtype, riptr->term_subtype);
		}
		if (riptr->term_subtype != NQS_NO_TERM) {
			if (db_flag > 2) {
				Ndebug("nqs_sent(3): \tmaking new "
					"segment node.\n");
			}
			make_segment(rqptr, AUX_REQ);
			riptr = rqptr->pplast->riptr;
			init_nqsbs(riptr, nbuf);
		}
	}

	if (nbuf->subtype == NQ_SENT_INIT) {
		riptr->init_subtype = NQ_SENT_INIT;

	} else if (nbuf->subtype == NQ_SENT_TERM) {
		/*
		 *	Reset the queue name to be the queue
		 *	the request will execute in, instead of 
		 *	the batch queue.
		 */
		strcpy(riptr->quename, nbuf->quename);
		if (riptr->init_subtype != NQ_SENT_INIT) {
			riptr->init_subtype = NQ_SENT_INIT;
		}
		riptr->utime = nbuf->utime;
		riptr->stime = nbuf->stime;
		riptr->term_subtype = NQ_SENT_TERM;

	} else {
		acct_err(ACCT_CAUT,
		       _("An unknown NQ_SENT subtype (%d) was found in the '%s' routine."),
			nbuf->subtype,
			"nqs_sent()");
	}

	return;
}


/*
 *	Process NQ_TERM records
 *	  Add the termination code to the seqinfo node.  Also, restart
 *	  the queue wait time timer.
 */
void
nqs_term(struct nqsbs *nbuf)
{
	struct	nqsbs	*riptr;
	struct	nqsreq	*rqptr;

	if (db_flag > 8) {
		Ndebug("nqs_term(9): TERM - seqno(%lld), mid(%lld).\n",
			nbuf->seqno, nbuf->mid);
	}

	if ((rqptr = find_request(nbuf->seqno, nbuf->mid)) == NULL) {
		make_request(NORM_REQ);
		rqptr = rhead->last;
		rqptr->seqno = nbuf->seqno;
		rqptr->mid = nbuf->mid;
		riptr = rqptr->slast->riptr;
		init_nqsbs(riptr, nbuf);

	} else if (rqptr->slast == NULL) {
		make_segment(rqptr, NORM_REQ);
		riptr = rqptr->slast->riptr;
		init_nqsbs(riptr, nbuf);

	} else {
		riptr = rqptr->slast->riptr;
	}

	/*
	 *	The last node on the segment list should be
	 *	what we want.  If not, the records in the file
	 *	are out of order.
	 */
	if (riptr->term_subtype == NQS_NO_TERM) {
		riptr->code = nbuf->code;
		riptr->utime = nbuf->utime;
		riptr->stime = nbuf->stime;
		riptr->term_subtype = nbuf->subtype;
		rqptr->sqtime = riptr->tmtime = nbuf->time;

	} else {
		if (db_flag > 2) {
			Ndebug("nqs_term(3): seqinfo node not "
				"at end of segment list.\n"
				"\triptr: term_subtype(%d), jid(0x%llx), "
				"nbuf->jid(0x%llx), rqptr->ltype(%d).\n",
				riptr->term_subtype, riptr->jid,
				nbuf->jid, rqptr->ltype);
		}
		return;	/* ignore this record */
	}

	rqptr->ltype    = nbuf->type;
	rqptr->lsubtype = nbuf->subtype;

	return;
}


/*
 *	Process NQ_SPOOL records
 *	  Update or create a nqsbs(c) entry on the pp list.  These
 *	  entries contain information about the pipeclients and
 *	  netclients.
 *
 */
void
nqs_spool(struct nqsbs *nbuf)
{
	struct	nqsbs	*riptr;
	struct	nqsreq	*rqptr;
	struct	nqsseg	*sptr;

	if (db_flag > 8) {
		Ndebug("nqs_spool(9): SPOOL - seqno(%lld), mid(%lld).\n",
			nbuf->seqno, nbuf->mid);
	}

	if ((rqptr = find_request(nbuf->seqno, nbuf->mid)) == NULL) {
		/*
		 * This is a new seqno. Create request, segment
		 * and nqsbs(c) entries for it.
		 */
		if (db_flag > 2) {
			Ndebug("nqs_spool(3): making request node for "
				"seqno(%lld), type NQ_SPOOL, subtype(%d).\n",
				nbuf->seqno, nbuf->subtype);
		}
		make_request(AUX_REQ);
		rqptr = rhead->last;	
		rqptr->seqno = nbuf->seqno;
		rqptr->mid = nbuf->mid;
		riptr = rqptr->pplast->riptr;
		init_nqsbs(riptr, nbuf);

	} else if (rqptr->pplast == NULL) {
		/*
		 * This seqno does not have any nqsbs(c) entries on the
		 * pp list, so make one.
		 */
		if (db_flag > 2) {
			Ndebug("nqs_spool(3): pplast == NULL.\n");
		}
		make_segment(rqptr, AUX_REQ);
		riptr = rqptr->pplast->riptr;
		init_nqsbs(riptr, nbuf);

	} else {
		/*
		 * Look through the entire pp list for this jid, since
		 * there could be multiple active pipeclients and
		 * netclients.  If jid = 0 (which could happen on a
		 * NQ_SPOOL_TERM record), then update the first active
		 * entry.
		 */
		int	found = 0;

		if (db_flag > 2) {
			Ndebug("nqs_spool(3):  looking for a nqsbs(c) "
				"entry with jid(0x%llx).\n", nbuf->jid);
		}

		for (sptr = rqptr->ppfirst, riptr = sptr->riptr;
		   ((sptr != NULL) && (found == 0)); sptr = sptr->nptr) {
			riptr = sptr->riptr;

			switch (nbuf->jid) {
			case 0:
				if (riptr->term_subtype == NQS_NO_TERM) {
					found = 1;
				}
				break;

			default:
				if (riptr->jid == nbuf->jid) {
					found = 1;
				}
				break;
			}		/* end of switch(jid) */
		}

		if (!found) {
			if (db_flag > 2) {
				Ndebug("nqs_spool(3): making new "
					"segment node.\n");
			}
			make_segment(rqptr, AUX_REQ);
			riptr = rqptr->pplast->riptr;
			init_nqsbs(riptr, nbuf);
		}
	}

	/*
	 * Update the nqsbs(c) entry.
	 */
	switch (nbuf->subtype) {

	case NQ_SPOOL_INIT:
		riptr->init_subtype = NQ_SPOOL_INIT;
		break;

	case NQ_SPOOL_TERM:
		if (riptr->init_subtype != NQ_SPOOL_INIT) {
			riptr->init_subtype = NQ_SPOOL_INIT;
		}
		riptr->utime = nbuf->utime;
		riptr->stime = nbuf->stime;
		riptr->term_subtype = NQ_SPOOL_TERM;
		break;

	default:
		acct_err(ACCT_CAUT,
		       _("An unknown NQ_SENT subtype (%d) was found in the '%s' routine."),
			nbuf->subtype,
			"nqs_spool()");
		break;
	}		/* end of switch(subtype) */

	return;
}


/*
 *	Dump the data structures.
 */
static void
dump_nqs_tbl()
{
	time_t	sqtime, stime;
	struct	nqsbs	*riptr;
	struct	nqsreq	*rqptr;
	struct	nqsseg	*sptr;

	rqptr = rhead->first;

	while (rqptr != NULL) {
		int	ltype = rqptr->ltype < 0 ? 0 : rqptr->ltype;
		int	lsubtype = rqptr->lsubtype < 0 ? 0 : rqptr->lsubtype;
		char	*lst = ltype == NQ_INIT ? Init_statname[lsubtype] :
				Term_statname[lsubtype];
		/*
		 *	Output the request node.
		 */
		sqtime = rqptr->sqtime;
		stime  = rqptr->stime;

		Ndebug("\nRequest node:\n");
		Ndebug("\tseqno %lld,  mid %lld,  last type(%d, %s), "
			"last subtype(%d, %s)\n\tstime:  %s\tsqtime: %s\n",
			rqptr->seqno, rqptr->mid,
			rqptr->ltype, Type_statname[ltype],
			rqptr->lsubtype, lst,
			ctime(&stime), ctime(&sqtime));

		/*
		 *	Output the request records.
		 */
		sptr = rqptr->sfirst;	/* NORM_REQ */
		Ndebug("  Normal requests\n");
		while (sptr != NULL) {
			riptr = sptr->riptr;
			print_nqs_rec(riptr);
			sptr = sptr->nptr;
		}

		/*
		 *	Output the Auxiliary records.
		 */
		sptr = rqptr->ppfirst;	/* AUX_REQ */
		Ndebug("  Aux requests\n");
		while (sptr != NULL) {
			riptr = sptr->riptr;
			print_nqs_rec(riptr);
			sptr = sptr->nptr;
		}

		rqptr = rqptr->nptr;
	}
	Ndebug("\n");

	return;
}


/*
 * ============================ tracknqs() ============================
 *	Keep track of all NQS records so we can condense them into a
 *	single record per user for writing to the sorted pacct file.
 */
void
tracknqs(struct nqsbs *nq, int uind)
{
	struct nqs_track *nt;
	int64_t	seqno, mid;
	uint64_t 	jid;

	/*
	 *	Get keys.
	 */
	seqno = nq->seqno;
	mid = nq->mid;
	jid = nq->jid;

	/*
	 *	Do we already have this one?  If so, keep track of
	 *	how many times we have already seen it and return.
	 */
	for(nt = tr_nqs; nt != NULL; nt = nt->nt_next) {
		if ((nt->nt_seqno == seqno) && (nt->nt_mid == mid) &&
		    (nt->nt_upind == uind) && (nt->nt_jid == jid) ) {
			(nt->nt_njob)++;
			if (db_flag > 4) {
				Ndebug("tracknqs(5): found seqno(%lld), "
					"mid(%lld), jid(0x%llx), upind(%d).\n",
					seqno, mid, jid, uind);
			}

			return;
		}
	}

	/*
	 *	Guess not, so add it to the front of the list.
	 */
	if ((nt = (struct nqs_track *)malloc(sizeof(struct nqs_track))) ==
			NULL) {
		acct_perr(ACCT_ABORT, errno,
			_("There was insufficient memory available when allocating '%s'."),
			"NQS Job");
	}

	/*
	 *	Fill it in.
	 */
	nt->nt_seqno = seqno;
	nt->nt_mid   = mid;
	nt->nt_upind = uind;
	nt->nt_jid   = jid;
	nt->nt_njob  = 0;

	/*
	 *	set up links.
	 */
	nt->nt_next = tr_nqs;
	tr_nqs = nt;

	if (db_flag > 4) {
		Ndebug("tracknqs(5): new node, - seqno(%lld), mid(%lld), "
			"jid(0x%llx), upind(%d).\n", seqno, mid, jid, uind);
	}

	return;
}

/*
 * ============================ seqcmp() ============================
 *	Compare routine for sorting nqs records by 
 *	seqno/mid/upind/jid.
 */
int
seqcmp(const void *e1, const void *e2)
{
	struct nqs_track *n1;
	struct nqs_track *n2;
	int64_t tmp;

	n1 = (struct nqs_track *)e1;
	n2 = (struct nqs_track *)e2;
	tmp = n1->nt_seqno - n2->nt_seqno;
	if (tmp == 0) {
		tmp = n1->nt_mid - n2->nt_mid;
		if (tmp == 0) {
			tmp = n1->nt_upind - n2->nt_upind;
			if (tmp == 0) {
				return(n1->nt_jid - n2->nt_jid);

			} else {
				return(tmp);
			}

		} else {
			return(tmp);
		}

	} else {
		return(tmp);
	}
}

/*
 * ============================ jidcmp() ============================
 *	Compare routine for sorting nqs records by jid.
 */
static int
jidcmp(const void *e1, const void *e2)
{
	struct nqs_track *n1;
	struct nqs_track *n2;

	n1 = (struct nqs_track *)e1;
	n2 = (struct nqs_track *)e2;
	
	if (n1->nt_jid < n2->nt_jid) {
		return(1);
	}

	return(-1);
}

struct nqs_track *nnt;
/*
 * ============================ Condense_nqs() ============================
 *	Put jobs together if they have common NQS records...
 *
 *	(Forget about combining records which came before first input uptime.
 *	Zap these so we do not mess up everything else.  -  this is wrong?)
 *
 *	The sequence number and machine ID make a unique key, so
 *	find all records with that unique key and then combine any
 *	associated jobs.
 */
void
Condense_nqs() 
{
	struct nqs_track *nt;
	int tablesz ;
	int ind, jnd;
	uint64_t jid;
	int upind, keepit, beginrec, size;
	int64_t seq, mid;
	int hindex;
	struct segment **htable;
	struct	segment *first_sptr, *next_sptr, *sptr;



	if (tr_nqs == NULL) {
		if (db_flag > 0) {
			Ndebug("Condense_nqs(1): No nqs jobs found.\n");
		}
		return;
	}

	/*
	 *	First, sort the NQS records by seq num, mid, and upind.
	 *	Start by building array nnt[] from the nqs linked list.
	 */
	for(ind = 0, nt = tr_nqs; nt != NULL; ind++,nt = nt->nt_next);
	tablesz = ind;

	size = sizeof(struct nqs_track) * tablesz;
	if ((nnt = (struct nqs_track *)malloc(size)) == NULL) {
		acct_perr(ACCT_ABORT, errno,
			_("There was insufficient memory available when allocating '%s'."),
			"final NQS table");
	}

	/*
	 *	Transfer data from the linked list to table nnt[]
	 */
	for(ind = 0, nt = tr_nqs; nt != NULL; ind++, nt = nt->nt_next) {
		nnt[ind].nt_seqno = nt->nt_seqno;
		nnt[ind].nt_mid   = nt->nt_mid;
		nnt[ind].nt_upind = nt->nt_upind;
		nnt[ind].nt_jid   = nt->nt_jid;
		nnt[ind].nt_njob  = nt->nt_njob;
		/*
		 *	Get rid of all records prior to uptime 0(?)
		 */
/*
 *		if ((Uptime_shifted) && (nnt[ind].nt_upind == 0)) {
 *			ind--;
 *		}
*/
	}
	tablesz = ind;
	Dump_nqs_tbl(nnt, tablesz, "DUMP NQS table before sort (nnt)");

	/*
	 *	Sort the table by seqno, mid, and upind
	 */
	qsort((char *)nnt, tablesz, sizeof(struct nqs_track), seqcmp);

	Dump_nqs_tbl(nnt, tablesz,
		"DUMP NQS table after sorting by seqno, mid, upind");

	/*
	 *	Now search the table and  zap records which do not matter
	 */
	for(ind = 0; ind < tablesz;) {
		seq = nnt[ind].nt_seqno;
		mid = nnt[ind].nt_mid;
		jid = nnt[ind].nt_jid;
		upind = nnt[ind].nt_upind;
		/*
		 *	Keep only the records which have (1) the same seqno/mid
		 *	and different upind or jid or (2) multiple segments for
		 *	the same jid/seqno/mid/upind.  Zap everything else.
		 */
		keepit = 0;
		beginrec = ind;
		while(seq == nnt[ind].nt_seqno && mid == nnt[ind].nt_mid) {
			if ((upind != nnt[ind].nt_upind) ||
			    (nnt[ind].nt_njob > 0) || (jid != nnt[ind].nt_jid)) {
				keepit = 1;
			}
			if (++ind == tablesz) {
				break;
			}
		}
		if (keepit == 0) {
			for(; beginrec < ind; beginrec++) {
				nnt[beginrec].nt_jid = ZAPJID;
			}
		}
	}

	/*
	 *	Now get rid of all entries which were zapped (jid = ZAPJID).
	 *	Sort nnt[] so zapped entries are at the end of the table
	 *	and then shorten the table.
	 */
	Dump_nqs_tbl(nnt, tablesz,
		"DUMP NQS table after zapping entries");
	qsort((char *)nnt, tablesz, sizeof(struct nqs_track), jidcmp);
	Dump_nqs_tbl(nnt, tablesz,
		"DUMP NQS table after sorting jid == ZAPJID to end");

	for(ind = 0; ind < tablesz && nnt[ind].nt_jid != ZAPJID; ind++);
	if (db_flag > 2) {
		Ndebug("Condense_nqs(3): NQS culling reduced records from "
			"%d to %d.\n", tablesz, ind);
	}
	tablesz = ind;
	if (tablesz == 0) {
		if (db_flag > 2) {
			Ndebug("Condense_nqs(3): No NQS jobs to be joined - "
				"returning.\n");
		}
		return;
	}

	/*
	 *	Put back into key sort order (seqno/mid/upind)
	 */
	Dump_nqs_tbl(nnt, tablesz,
		"DUMP NQS table before resort into key order");
	qsort((char *)nnt, tablesz, sizeof(struct nqs_track), seqcmp);

	Dump_nqs_tbl(nnt, tablesz, "DUMP NQS final table");
	Dump_cont_nqs(nnt, tablesz, "DUMP continued NQS segments");

	/*
	 *	Ok, now we are ready to combine NQS jobs (finally)
	 */
	for(ind = 0; ind < tablesz; ind = jnd) {
		/*
		 *	Get a pointer to the first part of the NQS
		 *	request.  Handle the case where jid == NQS_NO_JID.
		 *	This can occur if we have a rejected or queued
		 *	job, or are missing NQS accounting data.
		 */
		if (nnt[ind].nt_jid != NQS_NO_JID) {
			htable = uptime[nnt[ind].nt_upind].up_seg;
			hindex = HASH_INDEX(nnt[ind].nt_jid);
			if((first_sptr = seg_hashtable(htable, hindex, 
			    nnt[ind].nt_jid,
			    nnt[ind].nt_upind)) == (struct segment *)NULL ) {
				/* shouldn't happen; exit */
				acct_err(ACCT_ABORT,
				       _("No segment was returned from the hash table in the '%s' routine."),
					"Condense_nqs/seg_hashtable()");
			}
		} else {
			first_sptr = uptime[nnt[ind].nt_upind].up_nqs;
		}
		
		/*
		 *	If job IDs wrapped, it may not be this sptr...
		 *	Or, if this is an NQS_NO_JID part, we may not
		 *	be pointing to the right segment.
		 */
		sptr = first_sptr;
		seq = nnt[ind].nt_seqno;
		mid = nnt[ind].nt_mid;
		if (db_flag > 3) {
			Ndebug("Condense_nqs(4): Search for seqno(%lld), "
				"mid(%lld).\n", seq, mid);
		}
		while(sptr != NULL && sptr != BADPOINTER) {
			if (db_flag > 3) {
				Ndebug("Condense_nqs(4): Sptr sg_seqno(%lld), "
					"sg_mid(%lld).\n", sptr->sg_seqno,
					sptr->sg_mid);
			}
			if ((sptr->sg_seqno == seq) && (sptr->sg_mid == mid) ) {
				break; /* this is it */
			}
			sptr = sptr->sg_ptr;
		}
		if (sptr == NULL || sptr == BADPOINTER) {
			acct_err(ACCT_ABORT,
			       _("Could not find NQS sequence number (seqno) %lld, machine ID (mid) %lld,\n   Job ID (job) 0x%llx, uptime index (upind) %d."),
				seq, mid,
				nnt[ind].nt_jid, nnt[ind].nt_upind);
		}
		if (db_flag > 3) {
			Ndebug("Condense_nqs(4): Found seqno(%lld), "
				"mid(%lld), jid(0x%llx).\n",
				sptr->sg_seqno, sptr->sg_mid, nnt[ind].nt_jid);
		}
		first_sptr = sptr;

		/*
		 *	Combine all portions of an NQS job which have the 
		 *	same jid/seqno/mid/upind.
		 */
		Consolidate_nqs(first_sptr, seq, mid, ind);

		/*
		 *	Now, link any other entries which have the same
		 *	seqno/mid.
		 */
		for(jnd = ind+1; jnd < tablesz; jnd++) {
			if ((seq != nnt[jnd].nt_seqno) ||
			    (mid != nnt[jnd].nt_mid) ) {
				break;
			}
			if (db_flag > 4) {
				Ndebug("Condense_nqs(5): looking for another "
					"part to seqno(%lld), mid(%lld), "
					"jid(0x%llx)\n",
					seq, mid, nnt[jnd].nt_jid);
			}

			/*
			 *	Get the pointer to the next part of this
			 *	NQS job.  Handle the case where
			 *	jid == NQS_NO_JID.
			 */
			if (nnt[jnd].nt_jid != NQS_NO_JID) {
				htable = uptime[nnt[jnd].nt_upind].up_seg;
				hindex = HASH_INDEX(nnt[jnd].nt_jid);
				if((next_sptr = seg_hashtable(htable, hindex,
				    nnt[jnd].nt_jid ,nnt[jnd].nt_upind)) ==
				    (struct segment *)NULL ) {
					/* shouldn't happen; exit */
					acct_err(ACCT_ABORT,
					       _("No segment was returned from the hash table in the '%s' routine."),
						"Condense_nqs/seg_hashtable()");
				}
			} else {
				next_sptr = uptime[nnt[jnd].nt_upind].up_nqs;
			}


			/*
			 * Again, if job IDs wrapped it may not be this sptr...
			 */
			sptr = next_sptr;
			if (db_flag > 3) {
				Ndebug("Condense_nqs(4): Search for "
					"seqno(%lld), mid(%lld).\n", seq, mid);
			}
			while((sptr != NULL) && (sptr != BADPOINTER)) {
				if ((sptr->sg_seqno == seq) &&
						(sptr->sg_mid == mid) ) {
					if (db_flag > 3) {
						Ndebug("Condense_nqs(4): Found "
							"seqno(%lld), mid(%lld), "
							"jid(0x%llx).\n",
							sptr->sg_seqno,
							sptr->sg_mid,
							nnt[jnd].nt_jid);
					}
					break; /* this is it */
				}
				sptr = sptr->sg_ptr;
			}
			if ((sptr == NULL) || (sptr == BADPOINTER) ) {
				acct_err(ACCT_ABORT,
				       _("Could not find NQS sequence number (seqno) %lld, machine ID (mid) %lld,\n   Job ID (job) 0x%llx, uptime index (upind) %d."),
					seq, mid,
					nnt[ind].nt_jid, nnt[ind].nt_upind);
			}
			next_sptr = sptr;

			/*
			 *	Make sure all the portions of this
			 *	seqno/jid/upind request are combined together
			 */
			Consolidate_nqs(next_sptr, seq, mid, jnd);

			/*
			 *	Insert second chain on end of first chain.
			 */
			*(first_sptr->sg_lproc) = next_sptr->sg_proc;
			first_sptr->sg_lproc = next_sptr->sg_lproc;
			*(first_sptr->sg_ltape) = next_sptr->sg_tape;
			first_sptr->sg_ltape = next_sptr->sg_ltape;
			*(first_sptr->sg_lnqs) = next_sptr->sg_nqs;
			first_sptr->sg_lnqs = next_sptr->sg_lnqs;
			*(first_sptr->sg_lwm) = next_sptr->sg_wm;
			first_sptr->sg_lwm = next_sptr->sg_lwm;

			/*
			 *	Delete "next" chain 
			 */
			if ((next_sptr->sg_ptr != NOJOBFOUND) &&
			    (next_sptr->sg_ptr != BADPOINTER) ) {
				*next_sptr = *next_sptr->sg_ptr;

			} else {
				next_sptr->sg_ptr   = NULL;
				next_sptr->sg_seqno = -1;
				next_sptr->sg_mid   = -1;
				next_sptr->sg_reqid = -1;
				next_sptr->sg_arrayid = -1;
				next_sptr->sg_stop  = 0;
				next_sptr->sg_start = 0;
				next_sptr->sg_proc  = NULL;
				next_sptr->sg_lproc = &(next_sptr->sg_proc);
				next_sptr->sg_tape  = NULL;
				next_sptr->sg_ltape = &(next_sptr->sg_tape);
				next_sptr->sg_nqs   = NULL;
				next_sptr->sg_lnqs  = &(next_sptr->sg_nqs);
				next_sptr->sg_wm    = NULL;
				next_sptr->sg_lwm   = &(next_sptr->sg_wm);
			}
		}		/* end of for(jnd) */
	}		/* end of for(ind) */

	if (db_flag > 0) {
		Ndebug("Condense_nqs(1): Done condensing NQS records.\n");
	}

	return;
}


/*
 * =========================== Consolidate_nqs() ===========================
 *	Combine all portions of the specified NQS job which have the same
 *	jid/seqno/mid/upind.
 */
void
Consolidate_nqs(struct segment *first_sptr, int64_t seq, int64_t mid, int ind)
{
	struct segment	*next_sptr;
	struct segment *prev_sptr;
	int		njob = 0;

	if (nnt[ind].nt_njob == 0) {
		return;
	}

	prev_sptr = first_sptr;
	next_sptr = first_sptr->sg_ptr;
	while((next_sptr != NULL) && (next_sptr != BADPOINTER) ) {
		if ((next_sptr->sg_seqno == seq) && (next_sptr->sg_mid == mid)) {
			njob++;

			/*
			 *	Insert second chain on end of first chain.
			 */
			*(first_sptr->sg_lproc) = next_sptr->sg_proc;
			first_sptr->sg_lproc = next_sptr->sg_lproc;
			*(first_sptr->sg_ltape) = next_sptr->sg_tape;
			first_sptr->sg_ltape = next_sptr->sg_ltape;
			*(first_sptr->sg_lnqs) = next_sptr->sg_nqs;
			first_sptr->sg_lnqs = next_sptr->sg_lnqs;
			*(first_sptr->sg_lwm) = next_sptr->sg_wm;
			first_sptr->sg_lwm = next_sptr->sg_lwm;

			prev_sptr->sg_ptr = next_sptr->sg_ptr;
			free((char *)next_sptr);
			next_sptr = prev_sptr->sg_ptr;

		} else {
			prev_sptr = next_sptr;
			next_sptr = next_sptr->sg_ptr;
		}
	}
 
	if (db_flag > 4) {
		Ndebug("Consolidate_nqs(5): seqno(%lld), mid(%lld), njob(%d), "
			"jid(0x%llx).\n", seq, mid, njob, nnt[ind].nt_jid);
	}

	if (njob != nnt[ind].nt_njob) {
		if (db_flag > 4) {
			Ndebug("Consolidate_nqs(5): Didn't find all portions "
				"of NQS seqno(%lld), mid(%lld), upind(%d)!!\n" 
				"\t njob(%d) != nt_njob(%d)\n", 
				seq, mid, nnt[ind].nt_upind,
				njob, nnt[ind].nt_njob);
		}
	} 

	return;
}

/*
 *	add_segs_nqs() - go through the list of consolidated records
 *	and attach these to an appropriate segment; needed to wait
 *	until now to do this because some NQS record types don't
 *	contain valid jids.
 */
void add_segs_nqs()
{
	struct nqsbs *riptr;
	struct nqsreq *rqptr;
	struct nqsseg *nqsptr;
	struct segment *sp;
	struct Crec *Cptr;

	rqptr = rhead->first;
	while (rqptr != NULL) {
		nqsptr = rqptr->sfirst;
		while (nqsptr != NULL) {	
			riptr = nqsptr->riptr;
			sp = get_segptr(REC_NQS, nqsptr->upind, riptr->jid,
				riptr->time, 0L, 0L);
			if (sp == NOJOBFOUND) {
				/* skip over this one. */ 
			} else {
				/* Insert record at end of the NQS chain */
				Total_crecs++;
				Cptr = get_crec(riptr, (struct wkmgmtbs *)NULL, 
						REC_NQS);
				*(sp->sg_lnqs) = Cptr;
				sp->sg_lnqs = &(Cptr->cp_next);
				tracknqs(riptr, nqsptr->upind);
				if ((sp->sg_seqno != -1) && 
				    (sp->sg_seqno != riptr->seqno)) {
					acct_err(ACCT_ABORT,
					       _("An invalid (or multiple) sequence number, machine ID, or Job ID\n   occurred in the file '%s%d'\n  near offset (%lld).\n   Job ID 0x%llx is associated with NQS requests ID's %lld and %lld."),
						pacct,
					   nqsptr->findex, nqsptr->pacct_offset,
					   riptr->jid, riptr->seqno,
					   sp->sg_seqno);
				} else {
				    sp->sg_seqno = riptr->seqno;
				    sp->sg_mid = riptr->mid;
				}
			}
			nqsptr = nqsptr->nptr;
		}
		nqsptr = rqptr->ppfirst;
		while (nqsptr != NULL) {	
			riptr = nqsptr->riptr;
			sp = get_segptr(REC_NQS, nqsptr->upind, riptr->jid,
				riptr->time, 0L, 0L);
			if (sp == NOJOBFOUND) {
				/* skip over this one. */ 
			} else {
				/* Insert record at end of the NQS chain */
				Total_crecs++;
				Cptr = get_crec(riptr, (struct wkmgmtbs *)NULL, 
						REC_NQS);
				*(sp->sg_lnqs) = Cptr;
				sp->sg_lnqs = &(Cptr->cp_next);
				tracknqs(riptr, nqsptr->upind);
				if ((sp->sg_seqno != -1) && 
				    (sp->sg_seqno != riptr->seqno)) {
					acct_err(ACCT_ABORT,
					       _("An invalid (or multiple) sequence number, machine ID, or Job ID\n   occurred in the file '%s%d'\n  near offset (%lld).\n   Job ID 0x%llx is associated with NQS requests ID's %lld and %lld."),
						pacct,
					   nqsptr->findex, nqsptr->pacct_offset,
					   riptr->jid, riptr->seqno,
					   sp->sg_seqno);
				} else {
				    sp->sg_seqno = riptr->seqno;
				    sp->sg_mid = riptr->mid;
				}
			}
			nqsptr = nqsptr->nptr;
		}
		rqptr = rqptr->nptr;
	}
}
